@@ -28,23 +28,49 @@ import (
 	"github.com/golang/protobuf/proto"
 	"github.com/google/go-cmp/cmp"
 	"github.com/google/go-cmp/cmp/cmpopts"
-
 	"google.golang.org/grpc"
-	"google.golang.org/grpc/credentials/google"
-	"google.golang.org/grpc/credentials/insecure"
 	"google.golang.org/grpc/credentials/tls/certprovider"
 	"google.golang.org/grpc/internal"
 	"google.golang.org/grpc/internal/envconfig"
 	"google.golang.org/grpc/xds/bootstrap"
-	"google.golang.org/grpc/xds/internal/xdsclient/xdsresource/version"
 
-	v2corepb "github.com/envoyproxy/go-control-plane/envoy/api/v2/core"
 	v3corepb "github.com/envoyproxy/go-control-plane/envoy/config/core/v3"
 	structpb "github.com/golang/protobuf/ptypes/struct"
 )
 
 var (
-	v2BootstrapFileMap = map[string]string{
+	v3BootstrapFileMap = map[string]string{
+		"serverFeaturesIncludesXDSV3": `
+		{
+			"node": {
+				"id": "ENVOY_NODE_ID",
+				"metadata": {
+				    "TRAFFICDIRECTOR_GRPC_HOSTNAME": "trafficdirector"
+			    }
+			},
+			"xds_servers" : [{
+				"server_uri": "trafficdirector.googleapis.com:443",
+				"channel_creds": [
+					{ "type": "google_default" }
+				],
+				"server_features" : ["xds_v3"]
+			}]
+		}`,
+		"serverFeaturesExcludesXDSV3": `
+		{
+			"node": {
+				"id": "ENVOY_NODE_ID",
+				"metadata": {
+				    "TRAFFICDIRECTOR_GRPC_HOSTNAME": "trafficdirector"
+			    }
+			},
+			"xds_servers" : [{
+				"server_uri": "trafficdirector.googleapis.com:443",
+				"channel_creds": [
+					{ "type": "google_default" }
+				]
+			}]
+		}`,
 		"emptyNodeProto": `
 		{
 			"xds_servers" : [{
@@ -115,7 +141,8 @@ var (
 				"channel_creds": [
 					{ "type": "not-google-default" },
 					{ "type": "google_default" }
-				]
+				],
+				"server_features": ["xds_v3"]
 			}]
 		}`,
 		"goodBootstrap": `
@@ -130,7 +157,8 @@ var (
 				"server_uri": "trafficdirector.googleapis.com:443",
 				"channel_creds": [
 					{ "type": "google_default" }
-				]
+				],
+				"server_features": ["xds_v3"]
 			}]
 		}`,
 		"multipleXDSServers": `
@@ -144,7 +172,8 @@ var (
 			"xds_servers" : [
 				{
 					"server_uri": "trafficdirector.googleapis.com:443",
-					"channel_creds": [{ "type": "google_default" }]
+					"channel_creds": [{ "type": "google_default" }],
+					"server_features": ["xds_v3"]
 				},
 				{
 					"server_uri": "backup.never.use.com:1234",
@@ -152,9 +181,7 @@ var (
 				}
 			]
 		}`,
-	}
-	v3BootstrapFileMap = map[string]string{
-		"serverDoesNotSupportsV3": `
+		"serverSupportsIgnoreResourceDeletion": `
 		{
 			"node": {
 				"id": "ENVOY_NODE_ID",
@@ -167,23 +194,7 @@ var (
 				"channel_creds": [
 					{ "type": "google_default" }
 				],
-				"server_features" : ["foo", "bar"]
-			}]
-		}`,
-		"serverSupportsV3": `
-		{
-			"node": {
-				"id": "ENVOY_NODE_ID",
-				"metadata": {
-				    "TRAFFICDIRECTOR_GRPC_HOSTNAME": "trafficdirector"
-			    }
-			},
-			"xds_servers" : [{
-				"server_uri": "trafficdirector.googleapis.com:443",
-				"channel_creds": [
-					{ "type": "google_default" }
-				],
-				"server_features" : ["foo", "bar", "xds_v3"]
+				"server_features" : ["ignore_resource_deletion", "xds_v3"]
 			}]
 		}`,
 	}
@@ -194,14 +205,6 @@ var (
 			},
 		},
 	}
-	v2NodeProto = &v2corepb.Node{
-		Id:                   "ENVOY_NODE_ID",
-		Metadata:             metadata,
-		BuildVersion:         gRPCVersion,
-		UserAgentName:        gRPCUserAgentName,
-		UserAgentVersionType: &v2corepb.Node_UserAgentVersion{UserAgentVersion: grpc.Version},
-		ClientFeatures:       []string{clientFeatureNoOverprovisioning, clientFeatureResourceWrapper},
-	}
 	v3NodeProto = &v3corepb.Node{
 		Id:                   "ENVOY_NODE_ID",
 		Metadata:             metadata,
@@ -209,45 +212,51 @@ var (
 		UserAgentVersionType: &v3corepb.Node_UserAgentVersion{UserAgentVersion: grpc.Version},
 		ClientFeatures:       []string{clientFeatureNoOverprovisioning, clientFeatureResourceWrapper},
 	}
-	nilCredsConfigV2 = &Config{
+	nilCredsConfigNoServerFeatures = &Config{
 		XDSServer: &ServerConfig{
 			ServerURI: "trafficdirector.googleapis.com:443",
-			Creds:     grpc.WithTransportCredentials(insecure.NewCredentials()),
-			CredsType: "insecure",
-			NodeProto: v2NodeProto,
-		},
-		ClientDefaultListenerResourceNameTemplate: "%s",
-	}
-	nonNilCredsConfigV2 = &Config{
-		XDSServer: &ServerConfig{
-			ServerURI: "trafficdirector.googleapis.com:443",
-			Creds:     grpc.WithCredentialsBundle(google.NewComputeEngineCredentials()),
-			CredsType: "google_default",
-			NodeProto: v2NodeProto,
+			Creds:     ChannelCreds{Type: "insecure"},
 		},
+		NodeProto: v3NodeProto,
 		ClientDefaultListenerResourceNameTemplate: "%s",
 	}
 	nonNilCredsConfigV3 = &Config{
 		XDSServer: &ServerConfig{
-			ServerURI:    "trafficdirector.googleapis.com:443",
-			Creds:        grpc.WithCredentialsBundle(google.NewComputeEngineCredentials()),
-			CredsType:    "google_default",
-			TransportAPI: version.TransportV3,
-			NodeProto:    v3NodeProto,
+			ServerURI:      "trafficdirector.googleapis.com:443",
+			Creds:          ChannelCreds{Type: "google_default"},
+			ServerFeatures: []string{"xds_v3"},
 		},
+		NodeProto: v3NodeProto,
+		ClientDefaultListenerResourceNameTemplate: "%s",
+	}
+	nonNilCredsConfigWithDeletionIgnored = &Config{
+		XDSServer: &ServerConfig{
+			ServerURI:              "trafficdirector.googleapis.com:443",
+			Creds:                  ChannelCreds{Type: "google_default"},
+			IgnoreResourceDeletion: true,
+			ServerFeatures:         []string{"ignore_resource_deletion", "xds_v3"},
+		},
+		NodeProto: v3NodeProto,
+		ClientDefaultListenerResourceNameTemplate: "%s",
+	}
+	nonNilCredsConfigNoServerFeatures = &Config{
+		XDSServer: &ServerConfig{
+			ServerURI: "trafficdirector.googleapis.com:443",
+			Creds:     ChannelCreds{Type: "google_default"},
+		},
+		NodeProto: v3NodeProto,
 		ClientDefaultListenerResourceNameTemplate: "%s",
 	}
 )
 
 func (c *Config) compare(want *Config) error {
-	if diff := cmp.Diff(c, want,
+	if diff := cmp.Diff(want, c,
 		cmpopts.EquateEmpty(),
-		cmp.AllowUnexported(ServerConfig{}),
 		cmp.Comparer(proto.Equal),
 		cmp.Comparer(func(a, b grpc.DialOption) bool { return (a != nil) == (b != nil) }),
 		cmp.Transformer("certproviderconfigstring", func(a *certprovider.BuildableConfig) string { return a.String() }),
 	); diff != "" {
-		return fmt.Errorf("diff: %v", diff)
+		return fmt.Errorf("unexpected diff in config (-want, +got):\n%s", diff)
 	}
 	return nil
 }
@@ -273,7 +282,6 @@ func setupBootstrapOverride(bootstrapFileMap map[string]string) func() {
 // This function overrides the bootstrap file NAME env variable, to test the
 // code that reads file with the given fileName.
 func testNewConfigWithFileNameEnv(t *testing.T, fileName string, wantError bool, wantConfig *Config) {
-	t.Helper()
 	origBootstrapFileName := envconfig.XDSBootstrapFileName
 	envconfig.XDSBootstrapFileName = fileName
 	defer func() { envconfig.XDSBootstrapFileName = origBootstrapFileName }()
@@ -314,9 +322,9 @@ func testNewConfigWithFileContentEnv(t *testing.T, fileName string, wantError bo
 	}
 }
 
-// TestNewConfigV2ProtoFailure exercises the functionality in NewConfig with
+// TestNewConfigV3ProtoFailure exercises the functionality in NewConfig with
 // different bootstrap file contents which are expected to fail.
-func TestNewConfigV2ProtoFailure(t *testing.T) {
+func TestNewConfigV3ProtoFailure(t *testing.T) {
 	bootstrapFileMap := map[string]string{
 		"empty":          "",
 		"badJSON":        `["test": 123]`,
@@ -380,11 +388,11 @@ func TestNewConfigV2ProtoFailure(t *testing.T) {
 	}
 }
 
-// TestNewConfigV2ProtoSuccess exercises the functionality in NewConfig with
+// TestNewConfigV3ProtoSuccess exercises the functionality in NewConfig with
 // different bootstrap file contents. It overrides the fileReadFunc by returning
 // bootstrap file contents defined in this test, instead of reading from a file.
-func TestNewConfigV2ProtoSuccess(t *testing.T) {
-	cancel := setupBootstrapOverride(v2BootstrapFileMap)
+func TestNewConfigV3ProtoSuccess(t *testing.T) {
+	cancel := setupBootstrapOverride(v3BootstrapFileMap)
 	defer cancel()
 
 	tests := []struct {
@@ -395,47 +403,23 @@ func TestNewConfigV2ProtoSuccess(t *testing.T) {
 			"emptyNodeProto", &Config{
 				XDSServer: &ServerConfig{
 					ServerURI: "trafficdirector.googleapis.com:443",
-					Creds:     grpc.WithTransportCredentials(insecure.NewCredentials()),
-					CredsType: "insecure",
-					NodeProto: &v2corepb.Node{
-						BuildVersion:         gRPCVersion,
-						UserAgentName:        gRPCUserAgentName,
-						UserAgentVersionType: &v2corepb.Node_UserAgentVersion{UserAgentVersion: grpc.Version},
-						ClientFeatures:       []string{clientFeatureNoOverprovisioning, clientFeatureResourceWrapper},
-					},
+					Creds:     ChannelCreds{Type: "insecure"},
+				},
+				NodeProto: &v3corepb.Node{
+					UserAgentName:        gRPCUserAgentName,
+					UserAgentVersionType: &v3corepb.Node_UserAgentVersion{UserAgentVersion: grpc.Version},
+					ClientFeatures:       []string{clientFeatureNoOverprovisioning, clientFeatureResourceWrapper},
 				},
 				ClientDefaultListenerResourceNameTemplate: "%s",
 			},
 		},
-		{"unknownTopLevelFieldInFile", nilCredsConfigV2},
-		{"unknownFieldInNodeProto", nilCredsConfigV2},
-		{"unknownFieldInXdsServer", nilCredsConfigV2},
-		{"multipleChannelCreds", nonNilCredsConfigV2},
-		{"goodBootstrap", nonNilCredsConfigV2},
-		{"multipleXDSServers", nonNilCredsConfigV2},
-	}
-
-	for _, test := range tests {
-		t.Run(test.name, func(t *testing.T) {
-			testNewConfigWithFileNameEnv(t, test.name, false, test.wantConfig)
-			testNewConfigWithFileContentEnv(t, test.name, false, test.wantConfig)
-		})
-	}
-}
-
-// TestNewConfigV3Support verifies bootstrap functionality involving support for
-// the xDS v3 transport protocol. Here the client ends up using v2 or v3 based
-// on what the server supports.
-func TestNewConfigV3Support(t *testing.T) {
-	cancel := setupBootstrapOverride(v3BootstrapFileMap)
-	defer cancel()
-
-	tests := []struct {
-		name       string
-		wantConfig *Config
-	}{
-		{"serverDoesNotSupportsV3", nonNilCredsConfigV2},
-		{"serverSupportsV3", nonNilCredsConfigV3},
+		{"unknownTopLevelFieldInFile", nilCredsConfigNoServerFeatures},
+		{"unknownFieldInNodeProto", nilCredsConfigNoServerFeatures},
+		{"unknownFieldInXdsServer", nilCredsConfigNoServerFeatures},
+		{"multipleChannelCreds", nonNilCredsConfigV3},
+		{"goodBootstrap", nonNilCredsConfigV3},
+		{"multipleXDSServers", nonNilCredsConfigV3},
+		{"serverSupportsIgnoreResourceDeletion", nonNilCredsConfigWithDeletionIgnored},
 	}
 
 	for _, test := range tests {
@@ -449,21 +433,22 @@ func TestNewConfigV3Support(t *testing.T) {
 // TestNewConfigBootstrapEnvPriority tests that the two env variables are read
 // in correct priority.
 //
-// the case where the bootstrap file
-// environment variable is not set.
+// "GRPC_XDS_BOOTSTRAP" which specifies the file name containing the bootstrap
+// configuration takes precedence over "GRPC_XDS_BOOTSTRAP_CONFIG", which
+// directly specifies the bootstrap configuration in itself.
 func TestNewConfigBootstrapEnvPriority(t *testing.T) {
 	oldFileReadFunc := bootstrapFileReadFunc
 	bootstrapFileReadFunc = func(filename string) ([]byte, error) {
-		return fileReadFromFileMap(v2BootstrapFileMap, filename)
+		return fileReadFromFileMap(v3BootstrapFileMap, filename)
 	}
 	defer func() { bootstrapFileReadFunc = oldFileReadFunc }()
 
-	goodFileName1 := "goodBootstrap"
-	goodConfig1 := nonNilCredsConfigV2
+	goodFileName1 := "serverFeaturesIncludesXDSV3"
+	goodConfig1 := nonNilCredsConfigV3
 
-	goodFileName2 := "serverSupportsV3"
+	goodFileName2 := "serverFeaturesExcludesXDSV3"
 	goodFileContent2 := v3BootstrapFileMap[goodFileName2]
-	goodConfig2 := nonNilCredsConfigV3
+	goodConfig2 := nonNilCredsConfigNoServerFeatures
 
 	origBootstrapFileName := envconfig.XDSBootstrapFileName
 	envconfig.XDSBootstrapFileName = ""
@@ -481,21 +466,33 @@ func TestNewConfigBootstrapEnvPriority(t *testing.T) {
 	// When one of them is set, it should be used.
 	envconfig.XDSBootstrapFileName = goodFileName1
 	envconfig.XDSBootstrapFileContent = ""
-	if c, err := NewConfig(); err != nil || c.compare(goodConfig1) != nil {
-		t.Errorf("NewConfig() = %v, %v, want: %v, %v", c, err, goodConfig1, nil)
+	c, err := NewConfig()
+	if err != nil {
+		t.Errorf("NewConfig() failed: %v", err)
+	}
+	if err := c.compare(goodConfig1); err != nil {
+		t.Error(err)
 	}
 
 	envconfig.XDSBootstrapFileName = ""
 	envconfig.XDSBootstrapFileContent = goodFileContent2
-	if c, err := NewConfig(); err != nil || c.compare(goodConfig2) != nil {
-		t.Errorf("NewConfig() = %v, %v, want: %v, %v", c, err, goodConfig1, nil)
+	c, err = NewConfig()
+	if err != nil {
+		t.Errorf("NewConfig() failed: %v", err)
+	}
+	if err := c.compare(goodConfig2); err != nil {
+		t.Error(err)
 	}
 
 	// Set both, file name should be read.
 	envconfig.XDSBootstrapFileName = goodFileName1
 	envconfig.XDSBootstrapFileContent = goodFileContent2
-	if c, err := NewConfig(); err != nil || c.compare(goodConfig1) != nil {
-		t.Errorf("NewConfig() = %v, %v, want: %v, %v", c, err, goodConfig1, nil)
+	c, err = NewConfig()
+	if err != nil {
+		t.Errorf("NewConfig() failed: %v", err)
+	}
+	if err := c.compare(goodConfig1); err != nil {
+		t.Error(err)
 	}
 }
 
@@ -582,7 +579,7 @@ func TestNewConfigWithCertificateProviders(t *testing.T) {
 				"channel_creds": [
 					{ "type": "google_default" }
 				],
-				"server_features" : ["foo", "bar", "xds_v3"]
+				"server_features" : ["xds_v3"]
 			}],
 			"certificate_providers": {
 				"unknownProviderInstance1": {
@@ -608,7 +605,7 @@ func TestNewConfigWithCertificateProviders(t *testing.T) {
 				"channel_creds": [
 					{ "type": "google_default" }
 				],
-				"server_features" : ["foo", "bar", "xds_v3"],
+				"server_features" : ["xds_v3"],
 			}],
 			"certificate_providers": {
 				"unknownProviderInstance": {
@@ -632,9 +629,9 @@ func TestNewConfigWithCertificateProviders(t *testing.T) {
 			"xds_servers" : [{
 				"server_uri": "trafficdirector.googleapis.com:443",
 				"channel_creds": [
-					{ "type": "google_default" }
+					{ "type": "insecure" }
 				],
-				"server_features" : ["foo", "bar", "xds_v3"]
+				"server_features" : ["xds_v3"]
 			}],
 			"certificate_providers": {
 				"unknownProviderInstance": {
@@ -662,14 +659,20 @@ func TestNewConfigWithCertificateProviders(t *testing.T) {
 	cancel := setupBootstrapOverride(bootstrapFileMap)
 	defer cancel()
 
+	// Cannot use xdstestutils.ServerConfigForAddress here, as it would lead to
+	// a cyclic dependency.
+	jsonCfg := `{
+		"server_uri": "trafficdirector.googleapis.com:443",
+		"channel_creds": [{"type": "insecure"}],
+		"server_features": ["xds_v3"]
+	}`
+	serverCfg, err := ServerConfigFromJSON([]byte(jsonCfg))
+	if err != nil {
+		t.Fatalf("Failed to create server config from JSON %s: %v", jsonCfg, err)
+	}
 	goodConfig := &Config{
-		XDSServer: &ServerConfig{
-			ServerURI:    "trafficdirector.googleapis.com:443",
-			Creds:        grpc.WithCredentialsBundle(google.NewComputeEngineCredentials()),
-			CredsType:    "google_default",
-			TransportAPI: version.TransportV3,
-			NodeProto:    v3NodeProto,
-		},
+		XDSServer: serverCfg,
+		NodeProto: v3NodeProto,
 		CertProviderConfigs: map[string]*certprovider.BuildableConfig{
 			"fakeProviderInstance": wantCfg,
 		},
@@ -758,12 +761,10 @@ func TestNewConfigWithServerListenerResourceNameTemplate(t *testing.T) {
 			name: "goodServerListenerResourceNameTemplate",
 			wantConfig: &Config{
 				XDSServer: &ServerConfig{
-					ServerURI:    "trafficdirector.googleapis.com:443",
-					Creds:        grpc.WithCredentialsBundle(google.NewComputeEngineCredentials()),
-					CredsType:    "google_default",
-					TransportAPI: version.TransportV2,
-					NodeProto:    v2NodeProto,
+					ServerURI: "trafficdirector.googleapis.com:443",
+					Creds:     ChannelCreds{Type: "google_default"},
 				},
+				NodeProto:                                 v3NodeProto,
 				ServerListenerResourceNameTemplate:        "grpc/server?xds.resource.listening_address=%s",
 				ClientDefaultListenerResourceNameTemplate: "%s",
 			},
@@ -826,7 +827,7 @@ func TestNewConfigWithFederation(t *testing.T) {
 					"xds_servers": [{
 						"server_uri": "td.com",
 						"channel_creds": [ { "type": "google_default" } ],
-						"server_features" : ["foo", "bar", "xds_v3"]
+						"server_features" : ["xds_v3"]
 					}]
 				}
 			}
@@ -908,23 +909,19 @@ func TestNewConfigWithFederation(t *testing.T) {
 			name: "good",
 			wantConfig: &Config{
 				XDSServer: &ServerConfig{
-					ServerURI:    "trafficdirector.googleapis.com:443",
-					Creds:        grpc.WithCredentialsBundle(google.NewComputeEngineCredentials()),
-					CredsType:    "google_default",
-					TransportAPI: version.TransportV2,
-					NodeProto:    v2NodeProto,
+					ServerURI: "trafficdirector.googleapis.com:443",
+					Creds:     ChannelCreds{Type: "google_default"},
 				},
+				NodeProto:                                 v3NodeProto,
 				ServerListenerResourceNameTemplate:        "xdstp://xds.example.com/envoy.config.listener.v3.Listener/grpc/server?listening_address=%s",
 				ClientDefaultListenerResourceNameTemplate: "xdstp://xds.example.com/envoy.config.listener.v3.Listener/%s",
 				Authorities: map[string]*Authority{
 					"xds.td.com": {
 						ClientListenerResourceNameTemplate: "xdstp://xds.td.com/envoy.config.listener.v3.Listener/%s",
 						XDSServer: &ServerConfig{
-							ServerURI:    "td.com",
-							Creds:        grpc.WithCredentialsBundle(google.NewComputeEngineCredentials()),
-							CredsType:    "google_default",
-							TransportAPI: version.TransportV3,
-							NodeProto:    v3NodeProto,
+							ServerURI:      "td.com",
+							Creds:          ChannelCreds{Type: "google_default"},
+							ServerFeatures: []string{"xds_v3"},
 						},
 					},
 				},
@@ -934,12 +931,10 @@ func TestNewConfigWithFederation(t *testing.T) {
 			name: "goodWithDefaultDefaultClientListenerTemplate",
 			wantConfig: &Config{
 				XDSServer: &ServerConfig{
-					ServerURI:    "trafficdirector.googleapis.com:443",
-					Creds:        grpc.WithCredentialsBundle(google.NewComputeEngineCredentials()),
-					CredsType:    "google_default",
-					TransportAPI: version.TransportV2,
-					NodeProto:    v2NodeProto,
+					ServerURI: "trafficdirector.googleapis.com:443",
+					Creds:     ChannelCreds{Type: "google_default"},
 				},
+				NodeProto: v3NodeProto,
 				ClientDefaultListenerResourceNameTemplate: "%s",
 			},
 		},
@@ -947,12 +942,10 @@ func TestNewConfigWithFederation(t *testing.T) {
 			name: "goodWithDefaultClientListenerTemplatePerAuthority",
 			wantConfig: &Config{
 				XDSServer: &ServerConfig{
-					ServerURI:    "trafficdirector.googleapis.com:443",
-					Creds:        grpc.WithCredentialsBundle(google.NewComputeEngineCredentials()),
-					CredsType:    "google_default",
-					TransportAPI: version.TransportV2,
-					NodeProto:    v2NodeProto,
+					ServerURI: "trafficdirector.googleapis.com:443",
+					Creds:     ChannelCreds{Type: "google_default"},
 				},
+				NodeProto: v3NodeProto,
 				ClientDefaultListenerResourceNameTemplate: "xdstp://xds.example.com/envoy.config.listener.v3.Listener/%s",
 				Authorities: map[string]*Authority{
 					"xds.td.com": {
@@ -968,12 +961,10 @@ func TestNewConfigWithFederation(t *testing.T) {
 			name: "goodWithNoServerPerAuthority",
 			wantConfig: &Config{
 				XDSServer: &ServerConfig{
-					ServerURI:    "trafficdirector.googleapis.com:443",
-					Creds:        grpc.WithCredentialsBundle(google.NewComputeEngineCredentials()),
-					CredsType:    "google_default",
-					TransportAPI: version.TransportV2,
-					NodeProto:    v2NodeProto,
+					ServerURI: "trafficdirector.googleapis.com:443",
+					Creds:     ChannelCreds{Type: "google_default"},
 				},
+				NodeProto: v3NodeProto,
 				ClientDefaultListenerResourceNameTemplate: "xdstp://xds.example.com/envoy.config.listener.v3.Listener/%s",
 				Authorities: map[string]*Authority{
 					"xds.td.com": {
@@ -997,23 +988,26 @@ func TestNewConfigWithFederation(t *testing.T) {
 }
 
 func TestServerConfigMarshalAndUnmarshal(t *testing.T) {
-	c := ServerConfig{
-		ServerURI:    "test-server",
-		Creds:        nil,
-		CredsType:    "test-creds",
-		TransportAPI: version.TransportV3,
+	jsonCfg := `{
+		"server_uri": "test-server",
+		"channel_creds": [{"type": "insecure"}],
+		"server_features": ["xds_v3"]
+	}`
+	origConfig, err := ServerConfigFromJSON([]byte(jsonCfg))
+	if err != nil {
+		t.Fatalf("Failed to create server config from JSON %s: %v", jsonCfg, err)
 	}
-
-	bs, err := json.Marshal(c)
+	bs, err := json.Marshal(origConfig)
 	if err != nil {
 		t.Fatalf("failed to marshal: %v", err)
 	}
-	var cUnmarshal ServerConfig
-	if err := json.Unmarshal(bs, &cUnmarshal); err != nil {
+
+	unmarshaledConfig := new(ServerConfig)
+	if err := json.Unmarshal(bs, unmarshaledConfig); err != nil {
 		t.Fatalf("failed to unmarshal: %v", err)
 	}
-	if diff := cmp.Diff(cUnmarshal, c); diff != "" {
-		t.Fatalf("diff (-got +want): %v", diff)
+	if diff := cmp.Diff(origConfig, unmarshaledConfig); diff != "" {
+		t.Fatalf("Unexpected diff in server config (-want, +got):\n%s", diff)
 	}
 }
 
